
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "../../src/initng.h"

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <assert.h>
#include <string.h>
#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_common.h"
#include "../../src/initng_plugin_hook.h"

#include "initng_pidfile.h"


s_entry PIDFILE = { "pid_file", STRING, 8, DAEMON_OPT };

static int check_for_pidfile(active_h * s);
static int check_on_kill(active_h * s, process_h * p);


/*
 * When trying to start a service, don't allow it to
 * be set RUNNING if there must exist a pidfile, and
 * pid entry has to be updated.
 */
int check_if_done(active_h * s, e_a_status status)
{
    assert(s);
    assert(s->name);

    S_;

    /* only check if pidfile exists, if we are trying to start something */
    if (status != RUNNING)
        return (TRUE);

    /* Make sure the service uses a pidfile */
    if (!active_db_is(&PIDFILE, s))
        return (TRUE);

    /* if pidfile is set, we say ok for this status change */
    if (check_for_pidfile(s) == TRUE)
    {
        D_("Pidfine found and new pid set.\n");
        return (TRUE);
    }


    D_(" service %s is not up, it have to wait for pidfile first!\n",
       s->name);

    /* An alarm, make sure that this function will be called again */
    alarm(1);

    /* Don't allow this change, this will leave this service STARTING */
    return (FALSE);
}

/*
 * Check if a pidfile exists, if it exists, update the 
 * pid in the active_db entry. and return TRUE
 */
static int check_for_pidfile(active_h * s)
{
    int fd = 0;
    int len = 0;
    int newpid = 0;
    char buf[21];
    char *pidfile;
    process_h *process = NULL;

    assert(s);
    assert(s->name);

    S_;
    if (s->type != TYPE_DAEMON)
    {
        F_("Only daemon can have a pidfile.\n");
        return (FALSE);
    }

    while_processes(process, s)
    {
        if (process->pt == T_DAEMON)
            break;

    }

    /* The start_process have to be */
    if (!process || process->pt != T_DAEMON)
    {
        F_("Where is the process?\n");
        return (FALSE);
    }

    pidfile = active_db_get_string(&PIDFILE, s);
    if (!pidfile)
    {
        F_("Where is the pidfile string?\n");
        return (FALSE);
    }
    D_("service %s uses pidfile \"%s\"\n", s->name, pidfile);

    pidfile = fix_variables(pidfile, s);
    if (!pidfile)
    {
        F_("Where is the pidfile string?\n");
        return (FALSE);
    }
    D_("Also known as \"%s\"\n", pidfile);

    /* open pid file */
    fd = open(pidfile, O_RDONLY);

    /* If we cant open pidfile, this is bad */
    if (fd == -1)
    {
        D_("Unable to open pidfile: %s, \"%s\", it might not be created yet.",
           pidfile, strerror(errno));
        free(pidfile);
        return (FALSE);
    }

    /* Read data from buffer */
    len = read(fd, buf, 20);
    close(fd);
    if (len < 1)
    {
        F_("Read 0 chars from %s, Its empty.\n", pidfile);
        free(pidfile);
        return (FALSE);
    }

    /* remove last newline */
    if (buf[len - 1] == '\n')
        buf[len - 1] = 0;
    else
        buf[len] = 0;

    /* Try to convert pid to int */
    newpid = atoi(buf);
    if (newpid < 2)
    {
        F_("Failed to convert pid %s (%s) to int\n", buf, pidfile);
        free(pidfile);
        return (FALSE);
    }

    /* Check if the old pid == newpid) */
    if (newpid == process->pid)
    {
        D_("The pid in pidfile, is the same as the pid we have.\n");
        free(pidfile);
        return (TRUE);
    }


    /* See if it's a valid pid */
    if (kill(newpid, 0) < 0 && (errno == ESRCH))
    {
        F_("Pid %d found ( in file %s ), But there exists no process with pid %d.\n", newpid, pidfile, newpid);
        free(pidfile);
        return (FALSE);
    }

    /* Finally set it!\n */
    D_("Pid updated for %s, from %d to %d.\n", s->name, process->pid, newpid);
    process->pid = newpid;

    free(pidfile);
    return (TRUE);
}

static int check_on_kill(active_h * s, process_h * p)
{
    assert(s);
    assert(s->name);
    assert(p);

    S_;

    /* make sure its the start process this is about */
    if (p->pt != T_DAEMON && s->type != TYPE_DAEMON)
        return (FALSE);

    /* make sure this service is launching, else we dont do anythinh */
    if (s->a_status != STARTING)
        return (FALSE);

    /* this will alter check_if_done abowe */
    mark_service(s, RUNNING);
    return (TRUE);
}

int module_init(const char *version)
{
    D_("module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module, is compiled for \"%s\" version, and initng is compiled on \"%s\" version, wont load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    initng_sdt_add(&PIDFILE);
    initng_add_hook(CONFIRM_ASTATUS_CHANGE, 10, &check_if_done);
    initng_add_hook(HANDLE_KILLED, 10, &check_on_kill);

    return (TRUE);
}

void module_unload(void)
{
    D_("module_unload();\n");
    initng_sdt_del(&PIDFILE);
    initng_del_hook(CONFIRM_ASTATUS_CHANGE, &check_if_done);
    initng_del_hook(HANDLE_KILLED, &check_on_kill);

}
