
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <sys/un.h>                         /* memmove() strcmp() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <linux/kd.h>                       /* KDSIGACCEPT */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdio.h>                          /* printf() */
#include <stdlib.h>                         /* free() exit() */
#include <sys/reboot.h>                     /* reboot() RB_DISABLE_CAD */
#include <assert.h>

#include "initng_global.h"
#include "initng_handler.h"
#include "initng_process_db.h"
#include "initng_execute.h"
#include "initng_toolbox.h"
#include "initng_main.h"
#include "initng_common.h"

#include "main.h"
#include "initng_service_cache.h"
#include "initng_active_db.h"
#include "initng_load_module.h"
#include "initng_plugin_callers.h"
#include "initng_fd.h"

#include "initng_kill_handler.h"
#include "initng_static_data_id.h"

static void handle_killed_daemon(active_h * killed_service,
                                 process_h * process);
static void handle_killed_start(active_h * killed_service,
                                process_h * process);
static void handle_killed_stop(active_h * killed_service,
                               process_h * process);

/* called when a process got killed, identify it, and make a call with a pointer to the process */
void handle_killed_by_pid(pid_t kpid, int r_code)
{
    /* The process that got killed */
    active_h *service = NULL;
    process_h *process, *safe = NULL;

    D_("handle_killed_by_pid(%i);\n", kpid);

    /* don't do anything on this pid */
    if (kpid <= 1)
        return;

    /* Look in process database for a match */
    if (!(service = active_db_find_by_pid(kpid)))
    {
        D_("handle_killed_by_pid(%i): No match!\n", kpid);
        return;
    }

    D_("handle_killed_by_pid(%i): found service \"%s\"...\n", kpid,
       service->name);

    while_processes_safe(process, safe, service)
    {
        /* make sure thats the right process */
        if (process->pid != kpid)
            continue;

        /* set r_code */
        process->r_code = r_code;

        /* Check if a plugin wants to override handle_killed behavior */
        if (initng_plugin_handle_killed(service, process))
        {
            D_("Plugin did handle this kill.\n");
            return;
        }

        /* if no plugin catches, we try default functions below */
        switch (process->pt)
        {
            case T_START:
                handle_killed_start(service, process);
                break;
            case T_STOP:
                handle_killed_stop(service, process);
                break;
            case T_DAEMON:
                handle_killed_daemon(service, process);
                break;
            default:
                F_("process %i died with unknown handler, freeing process!\n",
                   kpid);
                list_del(&process->list);
                process_db_free(process);
                break;
        }
        return;
    }

    F_("handle_killed_by_pid(%i): traced killed pid to a service but can't find servicetype!\n", kpid);
    return;
}


static void handle_killed_daemon(active_h * service, process_h * process)
{
    assert(service);
    assert(service->name);
    assert(process);

    D_("(%s): initial status: \"%s\".\n",
       service->name, active_db_get_status_string(service->a_status));

    /* Set the universal variable, that signalize that something happend */
    g.interrupt = TRUE;

    /* Notify readpipe plugins that we're dead and pipes will be closed */
    initng_process_readpipe(service, process, TRUE);

    /*
     * If the return code (for example "exit 1", in a bash script)
     * from the program, is bigger than 0, this commonly signalize
     * that something got wrong, print this as an error msg on screen
     */
    if (process->r_code > 0)
        F_(" daemon %s, Returned with exit %i.\n", service->name,
           process->r_code);

    /* free the process, or start_service will fail */
    list_del(&process->list);
    process_db_free(process);

    /* OK, now accept this service as STOPPED, set this status */
    mark_service(service, STOPPED);
}

/*
 * handle_killed_start(), description:
 *
 * A process dies, and kill_handler walks the active_db looking for a match
 * on that pid, if it founds a match for a active->start_process, this
 * function is called, with a pointer to that active_service to
 * handle the situation.
 */
static void handle_killed_start(active_h * service, process_h * process)
{
    assert(service);
    assert(service->name);
    assert(process);

    D_("handle_killed_start(%s): initial status: \"%s\".\n",
       service->name, active_db_get_status_string(service->a_status));

    /* Set the universal variable, that signalize that something happend */
    g.interrupt = TRUE;

    /* Notify readpipe plugins that we're dead and pipes will be closed */
    initng_process_readpipe(service, process, TRUE);

    /*
     * If the return code (for example "exit 1", in a bash script)
     * from the program, is bigger than 0, this commonly signalize
     * that something got wrong, print this as an error msg on screen
     */
    if (process->r_code > 0)
        F_(" start %s, Returned with exit %i.\n", service->name,
           process->r_code);

    /*
     * If service is stopping, ignore this signal
     */
    if (service->a_status != STARTING)
    {
        F_("Start exited!, and service is not marked starting!\n");
        return;
    }


    /*
     * Make sure r_code don't signal error (can be override by UP_ON_FAILURE.
     */
    if (process->r_code && !active_db_is(&UP_ON_FAILURE, service))
    {

        mark_service(service, FAIL_STARTING);
        list_del(&process->list);
        process_db_free(process);
        return;
    }

    /* OK! now service is DONE! */
    mark_service(service, DONE);

    /* free the process struct, to spare memory */
    list_del(&process->list);
    process_db_free(process);
}


/* to do when a stop action dies */
static void handle_killed_stop(active_h * service, process_h * process)
{
    assert(service);
    assert(service->name);
    assert(process);

    g.interrupt = TRUE;
    D_("(%s);\n", service->name);

    if (service->a_status != STOPPING)
    {
        F_("stop service died, but service is not status STOPPING!\n");
    }
    /* Notify readpipe plugins that we're dead and pipes will be closed */
    initng_process_readpipe(service, process, TRUE);

    /*
     * If the return code (for example "exit 1", in a bash script)
     * from the program, is bigger than 0, this commonly signalize
     * that something got wrong, print this as an error msg on screen
     */
    if (process->r_code > 0)
        F_(" stop %s, Returned with exit %i.\n", service->name,
           process->r_code);


    /* free stopped process data */
    list_del(&process->list);
    process_db_free(process);

    /* mark service stopped */
    mark_service(service, STOPPED);
}
