#include <CUnit/CUnit.h>
#include <CUnit/Basic.h>

#include <stdlib.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <signal.h>
#include <ctype.h>

#include "test_initng.h"

/********************************************************************
 * initng_toolbox.c setup
 ********************************************************************/
#include "../src/initng_global.h"

static int global_init(void)
{
    initng_new(test_argc, test_argv, test_env);
    return 0;
}

static int global_cleanup(void)
{
    initng_free();
    return 0;
}

#include "../src/initng_toolbox.h"

/* 
 * invoke initng_calloc()
 */
static void test_initng_calloc(void)
{
    char *s;

    NEW_TEST();
    TEST("test allocating..");
    s = initng_calloc(1, 101);
    CU_ASSERT(s != NULL);
    free(s);
}

/* 
 * invoke initng_realloc() to both grow and shrink a block
 */
static void test_initng_realloc(void)
{
    char *s;

    NEW_TEST();

    TEST(" calloc size 100");
    s = initng_calloc(1, 100);
    CU_ASSERT(s != NULL);

    TEST(" realloc size 50");
    s = initng_realloc(s, 50);
    CU_ASSERT(s != NULL);

    TEST(" realloc size 200");
    s = initng_realloc(s, 200);
    CU_ASSERT(s != NULL);
    free(s);
}

/*
 * invoke i_strdup() and verify that it works
 */
static void test_i_strdup(void)
{
    char *p;
    char *s;

    NEW_TEST();

    TEST(" strdup ");
    p = "the quick brown fox jumps over the lazy dog";
    s = i_strdup(p);
    CU_ASSERT(s != NULL);
    CU_ASSERT(strcmp(s, p) == 0);
    free(s);
}

/*
 * invoke i_strndup() and verify that it works
 */
static void test_i_strndup(void)
{
    char *p;
    char *s;

    NEW_TEST();

    TEST(" strndup, size 100");
    p = "the quick brown fox jumps over the lazy dog";
    s = i_strndup(p, 100);
    CU_ASSERT(s != NULL);
    CU_ASSERT(strcmp(s, p) == 0);
    free(s);

    TEST("strndup, size 3");
    s = i_strndup(p, 3);
    CU_ASSERT(s != NULL);
    CU_ASSERT(strcmp(s, "the") == 0);
    free(s);
}

/* 
 * We want to test the memory functions when they handles the out of 
 * memory condition.
 *
 * 1. reduce the maximum amount of memory usable for the process, so
 *    that the alloc fails
 * 2. set an alarm() and a signal handler that will increase the
 *    maximum amount of memory back to the original level, so that
 *    the function eventually gets memory and returns
 * 3. invoke function
 */

static struct rlimit rlim_save;

static void test_out_of_memory_remove_limit(int signum)
{
    INFO("Removing limit");
    if (setrlimit(RLIMIT_AS, &rlim_save) != 0)
    {
        fprintf(stderr, "%s:%d Error %d with setrlimit(); %s\n",
                __FILE__, __LINE__, errno, strerror(errno));
        exit(1);
    }
}

static void test_out_of_memory(void)
{
    static struct rlimit rlim;
    struct sigaction sa;
    char *s;
    char *p;
    char *q;
    struct timeval start;
    struct timeval end;
    int usec;

    NEW_TEST();

    INFO("get the current memory limit");
    if (getrlimit(RLIMIT_AS, &rlim_save) != 0)
    {
        fprintf(stderr, "%s:%d Error %d with getrlimit(); %s\n",
                __FILE__, __LINE__, errno, strerror(errno));
        exit(1);
    }

    INFO("set up the signal handler for the alarm signal");
    memset(&sa, 0, sizeof(struct sigaction));
    sa.sa_handler = test_out_of_memory_remove_limit;
    sa.sa_flags = SA_RESTART;
    if (sigaction(SIGALRM, &sa, NULL) != 0)
    {
        fprintf(stderr, "%s:%d Error %d with sigaction(); %s\n",
                __FILE__, __LINE__, errno, strerror(errno));
        exit(1);
    }

    INFO("set the memory limit to a very low value");
    rlim = rlim_save;
    rlim.rlim_cur = 1;
    if (setrlimit(RLIMIT_AS, &rlim) != 0)
    {
        fprintf(stderr, "%s:%d Error %d with setrlimit(); %s\n",
                __FILE__, __LINE__, errno, strerror(errno));
        exit(1);
    }

    TEST(" initng_calloc ");

    INFO("set an alarm, and then invoke calloc");
    gettimeofday(&start, NULL);
    alarm(1);
    s = initng_calloc(1, 1024 * 1024);
    gettimeofday(&end, NULL);

    TEST(" verify allocation worked ");
    CU_ASSERT(s != NULL);

    TEST(" make sure it actually took time and restarted ");
    usec = (end.tv_sec - start.tv_sec) * 1000000;
    usec += end.tv_usec - start.tv_usec;
    CU_ASSERT(usec > 1000000);

    TEST("initng_realloc");

    INFO(" set our limit again");
    if (setrlimit(RLIMIT_AS, &rlim) != 0)
    {
        fprintf(stderr, "%s:%d Error %d with setrlimit(); %s\n",
                __FILE__, __LINE__, errno, strerror(errno));
        exit(1);
    }

    INFO(" and try with realloc ");
    gettimeofday(&start, NULL);
    alarm(1);
    s = initng_realloc(s, 2 * 1024 * 1024);
    gettimeofday(&end, NULL);

    TEST(" verify resize worked ");
    CU_ASSERT(s != NULL);

    INFO(" make sure it actually took time and restarted ");
    usec = (end.tv_sec - start.tv_sec) * 1000000;
    usec += end.tv_usec - start.tv_usec;
    CU_ASSERT(usec > 1000000);

    TEST(" i_strdup ");

    INFO(" set our limit again ");
    if (setrlimit(RLIMIT_AS, &rlim) != 0)
    {
        fprintf(stderr, "%s:%d Error %d with setrlimit(); %s\n",
                __FILE__, __LINE__, errno, strerror(errno));
        exit(1);
    }

    INFO(" and try with strdup ");
    gettimeofday(&start, NULL);
    alarm(1);
    memset(s, 'x', 1024 * 1024);
    s[1024 * 1024] = '\0';
    p = i_strdup(s);
    gettimeofday(&end, NULL);

    TEST(" verify strdup worked ");
    CU_ASSERT(p != NULL);

    TEST(" make sure it actually took time and restarted ");
    usec = (end.tv_sec - start.tv_sec) * 1000000;
    usec += end.tv_usec - start.tv_usec;
    CU_ASSERT(usec > 1000000);

    TEST(" i_strndup ");

    INFO(" set our limit again ");
    if (setrlimit(RLIMIT_AS, &rlim) != 0)
    {
        fprintf(stderr, "%s:%d Error %d with setrlimit(); %s\n",
                __FILE__, __LINE__, errno, strerror(errno));
        exit(1);
    }

    INFO(" and try with strdup ");
    gettimeofday(&start, NULL);
    alarm(1);
    q = i_strndup(s, 1024 * 512);
    gettimeofday(&end, NULL);

    TEST(" verify strndup worked ");
    CU_ASSERT(q != NULL);

    TEST(" make sure it actually took time and restarted ");
    usec = (end.tv_sec - start.tv_sec) * 1000000;
    usec += end.tv_usec - start.tv_usec;
    CU_ASSERT(usec > 1000000);

    /* free memory */
    free(s);
    free(p);
    free(q);
}

/*
 * invoke set_proc_title() and verify that it works
 *
 * In order to do this, we use "ps" to find our own process id in the
 * process list, and look at the name there. This test depends on 
 * "ps" being available. It would probably be better to look at 
 * /proc/$pid, but this is easy enough.
 *
 * Note that this is not a very good test, as it depends on the amount
 * of space available for the process to set its own title. This is 
 * the amount of space allocated by the kernel for argv[] and env[],
 * which might be too little for the actual initng process.
 */

static char *get_process_name(void)
{
    FILE *fp;
    int pid;
    char line[1024];
    char *s;
    char *name;

    fp = popen("ps wwwwo pid,command", "r");
    if (fp == NULL)
    {
        fprintf(stderr, "%s:%d Error with popen()\n", __FILE__, __LINE__);
        exit(1);
    }
    pid = getpid();
    s = NULL;
    while (fgets(line, sizeof(line), fp) != NULL)
    {
        if (atoi(line) == pid)
        {
            s = line;
            while (isspace(*s))
                s++;                        /* skip leading whitespace */
            while (isdigit(*s))
                s++;                        /* skip pid */
            s++;                            /* skip space after pid */

            /* copy the name */
            name = strdup(s);
            if (name == NULL)
            {
                fprintf(stderr, "%s:%d Out of memory\n", __FILE__, __LINE__);
                exit(1);
            }
            /* find the trailing newline */
            s = strchr(name, '\n');
            /* back up removing whitespace, as shortened names have ' ' added */
            while ((s != name) && isspace(*s))
                s--;
            *(s + 1) = '\0';
        }
    }
    fclose(fp);
    if (s == NULL)
    {
        fprintf(stderr, "%s:%d My process not found\n", __FILE__, __LINE__);
        exit(1);
    }
    return name;
}

static void test_set_proc_title(void)
{
    char *org_proc_title;
    char *new_title;
    char *new_proc_title;

    NEW_TEST();
    TEST("proc_title");
    org_proc_title = get_process_name();
    new_title = "fubar";
    set_proc_title(new_title);
    new_proc_title = get_process_name();
    CU_ASSERT(strcmp(new_title, new_proc_title) == 0);
}


void test_initng_toolbox_setup(void)
{
    CU_pSuite suite;

    suite = CU_add_suite("initng_toolbox", global_init, global_cleanup);
    CU_add_test(suite, "initng_calloc", test_initng_calloc);
    CU_add_test(suite, "initng_realloc", test_initng_realloc);
    CU_add_test(suite, "initng_i_strdup", test_i_strdup);
    CU_add_test(suite, "initng_i_strndup", test_i_strndup);
    CU_add_test(suite, "out of memory", test_out_of_memory);
    CU_add_test(suite, "set_proc_title", test_set_proc_title);
}
